import json
import os
import random
import re
from datetime import timedelta
from threading import Thread

from nostr_sdk import Client, Timestamp, PublicKey, Tag, Keys, Options, SecretKey, NostrSigner, EventId, Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import post_process_list_to_users

"""
This File contains a Module to find inactive follows for a user on nostr

Accepted Inputs: None needed
Outputs: A list of users that have been inactive 
Params:  None
"""


class DiscoverInactiveFollows(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_PEOPLE_DISCOVERY
    TASK: str = "inactive-follows"
    FIX_COST: float = 50
    client: Client
    dvm_config: DVMConfig

    def __init__(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                 admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        super().__init__(name=name, dvm_config=dvm_config, nip89config=nip89config, nip88config=nip88config,
                         admin_config=admin_config, options=options)

    def is_input_supported(self, tags, client=None, dvm_config=None):
        # no input required
        return True

    def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config

        request_form = {"jobID": event.id().to_hex()}

        # default values
        user = event.author().to_hex()
        since_days = 90

        for tag in event.tags():
            if tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "user":  # check for param type
                    user = tag.as_vec()[2]
                elif param == "since_days":  # check for param type
                    since_days = int(tag.as_vec()[2])

        options = {
            "user": user,
            "since_days": since_days
        }
        request_form['options'] = json.dumps(options)
        return request_form

    def process(self, request_form):
        from nostr_sdk import Filter
        from types import SimpleNamespace
        ns = SimpleNamespace()

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)
        cli = Client.with_opts(signer, opts)
        for relay in self.dvm_config.RELAY_LIST:
            cli.add_relay(relay)
        cli.connect()

        options = DVMTaskInterface.set_options(request_form)



        inactivefollowerslist = ""
        relay_list = ["wss://relay.damus.io", "wss://nostr-pub.wellorder.net", "wss://nos.lol", "wss://nostr.wine",
                      "wss://relay.nostfiles.dev", "wss://nostr.mom", "wss://nostr.oxtr.dev",
                      "wss://relay.nostr.bg", "wss://relay.f7z.io"]
        relaytimeout = 5
        step = 20
        keys = Keys.parse(os.getenv(env.NOSTR_PRIVATE_KEY))
        opts = Options().wait_for_send(False).send_timeout(timedelta(seconds=5)).skip_disconnected_relays(
            True)
        cl = Client.with_opts(keys, opts)
        for relay in relay_list:
            cl.add_relay(relay)
        cl.connect()

        timeinseconds = 3 * 24 * 60 * 60
        dif = Timestamp.now().as_secs() - timeinseconds
        considernotessince = Timestamp.from_secs(dif)
        filt = Filter().author(user).kind(1).since(considernotessince)
        reactions = cl.get_events_of([filt], timedelta(seconds=relaytimeout))
        list = []
        random.shuffle(reactions)
        for reaction in reactions:
            if reaction.kind() == 1:
                list.append(reaction.content())
        all = json.dumps(list)
        all = all.replace("\n", " ").replace("\n\n", " ")
        cleared = ""
        tokens = all.split()
        for item in tokens:
            item = item.replace("\n", " ").lstrip("\"").rstrip(",").rstrip(("."))
            # print(item)
            if item.__contains__("http") or item.__contains__("\nhttp") or item.__contains__(
                    "\n\nhttp") or item.lower().__contains__("nostr:") or item.lower().__contains__(
                    "nevent") or item.__contains__("\\u"):
                cleareditem = ""
            else:
                cleareditem = item
            cleared = cleared + " " + cleareditem

        cleared = cleared.replace("\n", " ")
        # res = re.sub(r"[^ a-zA-Z0-9.!?/\\:,]+", '', all)
        # print(cleared)
        try:
            answer = LLAMA2(
                "Give me the 15 most important substantives as keywords of the following input: " + cleared,
                "nostruser",
                "Reply only with a comma-seperated keywords. return topics starting with a *", clear=True)
        except:
            answer = ""

        promptarr = answer.split(":")
        if len(promptarr) > 1:
            # print(promptarr[1])
            prompt = promptarr[1].lstrip("\n").replace("\n", ",").replace("*", ",").replace("•", ",")
        else:
            prompt = promptarr[0].replace("\n", ",").replace("*", "")

        pattern = r"[^a-zA-Z,#'\s]"
        text = re.sub(pattern, "", prompt) + ","

        # text = (text.replace("Let's,", "").replace("Why,", "").replace("GM,", "")
        #        .replace("Remember,", "").replace("I,", "").replace("Think,", "")
        #        .replace("Already,", ""))
        # print(text)
        keywords = text.split(', ')
        keywords = [x.lstrip().rstrip(',') for x in keywords if x]

        print(keywords)

        # answer = LLAMA2("Extent the given list with 5 synonyms per entry  " + str(keywords), user,
        #                "Reply only with a comma-seperated keywords. return topics starting with a *")
        # answer.replace(" - Alternatives:", ",")
        # print(answer)
        # promptarr = answer.split(":")
        # if len(promptarr) > 1:
        #    # print(promptarr[1])
        #    prompt = promptarr[1].lstrip("\n").replace("\n", ",").replace("*", "").replace("•", "")
        # else:
        #    prompt = promptarr[0].replace("\n", ",").replace("*", "")

        # pattern = r"[^a-zA-Z,'\s]"
        # text = re.sub(pattern, "", prompt) + ","
        # keywords = text.split(', ')

        # print(keywords)

        timeinseconds = 30 * 60  # last 30 min?
        dif = Timestamp.now().as_secs() - timeinseconds
        looksince = Timestamp.from_secs(dif)
        filt2 = Filter().kind(1).since(looksince)
        notes = cl.get_events_of([filt2], timedelta(seconds=6))

        # finallist = []
        ns.finallist = {}

        print("Notes found: " + str(len(notes)))

        def scanList(noteid: EventId, instance, i, length):

            relay_list = ["wss://relay.damus.io", "wss://nostr-pub.wellorder.net", "wss://nos.lol",
                          "wss://nostr.wine",
                          "wss://relay.nostfiles.dev", "wss://nostr.mom", "wss://nostr.oxtr.dev",
                          "wss://relay.nostr.bg", "wss://relay.f7z.io"]
            keys = Keys.parse(os.getenv(env.NOSTR_PRIVATE_KEY))
            opts = Options().wait_for_send(wait_for_send).send_timeout(
                timedelta(seconds=5)).skip_disconnected_relays(True)
            cli = Client.with_opts(keys, opts)
            for relay in relay_list:
                cli.add_relay(relay)
            cli.connect()

            filters = []
            instance.finallist[noteid.to_hex()] = 0
            filt = Filter().kinds([9735, 7, 1]).event(noteid)
            reactions = cl.get_events_of([filt], timedelta(seconds=5))
            print(str(len(reactions)) + "   " + str(j) + "/" + str(len(notes)))
            instance.finallist[noteid.to_hex()] = len(reactions)

            print(str(i) + "/" + str(length))
            cli.disconnect()

        j = 0

        threads = []

        for note in notes:

            j = j + 1
            res = [ele for ele in keywords if (ele.replace(',', "") in note.content())]
            if bool(res):
                if not note.id().to_hex() in ns.finallist and note.pubkey().to_hex() != user:
                    args = [note.id(), ns, j, len(notes)]
                    t = Thread(target=scanList, args=args)
                    threads.append(t)

            # Start all threads
        for x in threads:
            x.start()

            # Wait for all of them to finish
        for x in threads:
            x.join()

        finallist_sorted = sorted(ns.finallist.items(), key=lambda x: x[1], reverse=True)
        converted_dict = dict(finallist_sorted)
        print(json.dumps(converted_dict))

        notelist = ""
        resultlist = []
        i = 0
        notelist = "Based on topics: " + json.dumps(keywords).lstrip("[").rstrip(("]")) + "\n\n"
        for k in converted_dict:
            # print(k)
            if is_bot:
                i = i + 1
                notelist = notelist + "nostr:" + EventId.from_hex(k).to_bech32() + "\n\n"
                if i == 25:
                    break
            else:
                p_tag = Tag.parse(["p", k])
                resultlist.append(p_tag.as_vec())

        else:
            return json.dumps(resultlist[:25])

    def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_users(result)

        # if not text/plain, don't post-process
        return result


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/b29b6ec4bf9b6184f69d33cb44862db0d90a2dd9a506532e7ba5698af7d36210.jpg",
        "about": "I discover users you follow, but that have been inactive on Nostr",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "amount": "Free",
        "nip90Params": {
            "user": {
                "required": False,
                "values": [],
                "description": "Do the task for another user"
            },
            "since_days": {
                "required": False,
                "values": [],
                "description": "The number of days a user has not been active to be considered inactive"
            }
        }
    }
    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return DiscoverInactiveFollows(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                   admin_config=admin_config)


if __name__ == '__main__':
    process_venv(DiscoverInactiveFollows)



import json
import math
import os
import signal
import time
from datetime import timedelta

from nostr_sdk import (Keys, Client, Timestamp, Filter, nip04_decrypt, HandleNotification, EventBuilder, PublicKey,
                       Options, Tag, Event, nip04_encrypt, NostrSigner, EventId, Nip19Event, nip44_decrypt, Kind)

from nostr_dvm.utils.database_utils import fetch_user_metadata
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nip88_utils import nip88_has_active_subscription
from nostr_dvm.utils.nip89_utils import NIP89Config
from nostr_dvm.utils.nwc_tools import nwc_zap
from nostr_dvm.utils.subscription_utils import create_subscription_sql_table, add_to_subscription_sql_table, \
    get_from_subscription_sql_table, update_subscription_sql_table, get_all_subscriptions_from_sql_table, \
    delete_from_subscription_sql_table
from nostr_dvm.utils.zap_utils import create_bolt11_lud16, zaprequest


class Subscription:
    job_list: list

    # This is a simple list just to keep track which events we created and manage, so we don't pay for other requests
    def __init__(self, dvm_config, admin_config=None):
        self.NAME = "Subscription Handler"
        dvm_config.DB = "db/" + "subscriptions" + ".db"
        self.dvm_config = dvm_config
        nip89config = NIP89Config()
        nip89config.NAME = self.NAME
        self.dvm_config.NIP89 = nip89config
        self.admin_config = admin_config
        self.keys = Keys.parse(dvm_config.PRIVATE_KEY)
        wait_for_send = False
        skip_disconnected_relays = True
        opts = (Options().wait_for_send(wait_for_send).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT))
                .skip_disconnected_relays(skip_disconnected_relays))
        signer = NostrSigner.keys(self.keys)
        self.client = Client.with_opts(signer, opts)

        pk = self.keys.public_key()

        self.job_list = []

        print("Nostr Subscription Handler public key: " + str(pk.to_bech32()) + " Hex: " + str(
            pk.to_hex()) + "\n")

        for relay in self.dvm_config.RELAY_LIST:
            self.client.add_relay(relay)
        self.client.connect()

        zap_filter = Filter().pubkey(pk).kinds([EventDefinitions.KIND_ZAP]).since(Timestamp.now())
        cancel_subscription_filter = Filter().kinds([EventDefinitions.KIND_NIP88_STOP_SUBSCRIPTION_EVENT]).since(
            Timestamp.now())
        authors = []
        if admin_config is not None:
            for key in admin_config.USERNPUBS:
                authors.append(PublicKey.parse(key))
        dm_filter = Filter().authors(authors).pubkey(pk).kinds([EventDefinitions.KIND_DM]).since(Timestamp.now())

        self.client.subscribe([zap_filter, dm_filter, cancel_subscription_filter], None)

        create_subscription_sql_table(dvm_config.DB)


        class NotificationHandler(HandleNotification):
            client = self.client
            dvm_config = self.dvm_config
            keys = self.keys

            def handle(self, relay_url, subscription_id, nostr_event: Event):
                if nostr_event.kind().as_u64() == EventDefinitions.KIND_DM.as_u64():
                    print(nostr_event.as_json())
                    handle_dm(nostr_event)
                elif nostr_event.kind().as_u64() == EventDefinitions.KIND_NIP88_STOP_SUBSCRIPTION_EVENT.as_u64():
                    handle_cancel(nostr_event)

            def handle_msg(self, relay_url, msg):
                return

        def handle_cancel(nostr_event):
            print(nostr_event.as_json())
            sender = nostr_event.author().to_hex()
            kind7001eventid = ""
            recipient = ""
            if sender == self.keys.public_key().to_hex():
                return

            for tag in nostr_event.tags():
                if tag.as_vec()[0] == "p":
                    recipient = tag.as_vec()[1]
                elif tag.as_vec()[0] == "e":
                    kind7001eventid = tag.as_vec()[1]

            if kind7001eventid != "":
                subscription = get_from_subscription_sql_table(dvm_config.DB, kind7001eventid)

                if subscription is not None:
                    update_subscription_sql_table(dvm_config.DB, kind7001eventid, recipient,
                                                  subscription.subscriber, subscription.nwc, subscription.cadence,
                                                  subscription.amount, subscription.begin, subscription.end,
                                                  subscription.tier_dtag, subscription.zaps, subscription.recipe,
                                                  False, Timestamp.now().as_secs())

        def infer_subscription_end_time(start, cadence):
            end = start
            if cadence == "daily":
                end = start + 60 * 60 * 24
            elif cadence == "weekly":
                end = start + 60 * 60 * 24 * 7
            elif cadence == "monthly":
                # TODO check days of month -.-
                end = start + 60 * 60 * 24 * 31
            elif cadence == "yearly":
                # TODO check extra day every 4 years
                end = start + 60 * 60 * 24 * 356
            return end

        def pay_zap_split(nwc, overall_amount, zaps):
            overallsplit = 0

            for zap in zaps:
                overallsplit += int(zap['split'])

            zapped_amount = 0
            for zap in zaps:
                name, nip05, lud16 = fetch_user_metadata(zap['key'], self.client)
                splitted_amount = math.floor(
                    (int(zap['split']) / overallsplit) * int(overall_amount) / 1000)
                # invoice = create_bolt11_lud16(lud16, splitted_amount)
                # TODO add details about DVM in message
                invoice = zaprequest(lud16, splitted_amount, "DVM subscription", None,
                                     PublicKey.parse(zap['key']), self.keys, DVMConfig.RELAY_LIST)
                print(invoice)
                if invoice is not None:
                    nwc_event_id = nwc_zap(nwc, invoice, self.keys, zap['relay'])
                    if nwc_event_id is None:
                        print("error zapping " + lud16)
                    else:
                        zapped_amount = zapped_amount + (splitted_amount * 1000)
                        print(str(zapped_amount) + "/" + str(overall_amount))

            if zapped_amount < overall_amount * 0.8:  # TODO how do we handle failed zaps for some addresses? we are ok with 80% for now
                success = False
            else:
                print("Zapped successfully")
                success = True
                # if no active subscription exists OR the subscription ended, pay

            return success

        def make_subscription_zap_recipe(event7001, recipient, subscriber, start, end, tier_dtag):
            message = "payed by subscription service"
            pTag = Tag.parse(["p", recipient])
            PTag = Tag.parse(["P", subscriber])
            eTag = Tag.parse(["e", event7001])
            validTag = Tag.parse(["valid", str(start), str(end)])
            tierTag = Tag.parse(["tier", tier_dtag])
            alttag = Tag.parse(["alt", "This is a NIP90 DVM Subscription Payment Recipe"])

            tags = [pTag, PTag, eTag, validTag, tierTag, alttag]

            event = EventBuilder(EventDefinitions.KIND_NIP88_PAYMENT_RECIPE,
                                 message, tags).to_event(self.keys)

            dvmconfig = DVMConfig()
            signer = NostrSigner.keys(self.keys)
            client = Client(signer)
            for relay in dvmconfig.RELAY_LIST:
                client.add_relay(relay)
            client.connect()
            recipeid = client.send_event(event)
            recipe = recipeid.to_hex()
            return recipe

        def handle_dm(nostr_event):

            sender = nostr_event.author().to_hex()
            if sender == self.keys.public_key().to_hex():
                return

            try:
                decrypted_text = nip04_decrypt(self.keys.secret_key(), nostr_event.author(), nostr_event.content())
                try:
                    jsonevent = json.loads(decrypted_text)
                    print(jsonevent)
                    nwc = nip44_decrypt(self.keys.secret_key(), nostr_event.author(), jsonevent['nwc'])
                    event7001 = jsonevent['subscribe_event']
                    cadence = jsonevent['cadence']
                    recipient = jsonevent['recipient']
                    subscriber = jsonevent['subscriber']
                    overall_amount = int(jsonevent['overall_amount'])
                    tier_dtag = jsonevent['tier_dtag']

                    start = Timestamp.now().as_secs()

                    isactivesubscription = False
                    recipe = ""

                    subscription = get_from_subscription_sql_table(dvm_config.DB, event7001)
                    #if subscription is not None and subscription.end > start:
                    #    start = subscription.end
                    #    isactivesubscription = True



                    zapsstr = json.dumps(jsonevent['zaps'])
                    print(zapsstr)
                    success = True
                    if subscription is None or subscription.end <= Timestamp.now().as_secs():
                        #rather check nostr if our db is right
                        subscription_status = nip88_has_active_subscription(
                            PublicKey.parse(subscriber),
                            tier_dtag, self.client, recipient, checkCanceled=False)

                        if not subscription_status["isActive"]:
                            success = pay_zap_split(nwc, overall_amount, jsonevent['zaps'])
                            start = Timestamp.now().as_secs()
                            end = infer_subscription_end_time(start, cadence)
                        else:
                            start = Timestamp.now().as_secs()
                            end = subscription_status["validUntil"]
                    else:
                        start = subscription.begin
                        end = subscription.end


                    if success:
                        recipe = make_subscription_zap_recipe(event7001, recipient, subscriber, start, end, tier_dtag)
                        print("RECIPE " + recipe)
                        isactivesubscription = True

                    if subscription is None:
                        add_to_subscription_sql_table(dvm_config.DB, event7001, recipient, subscriber, nwc,
                                                      cadence, overall_amount, start, end, tier_dtag,
                                                      zapsstr, recipe, isactivesubscription, Timestamp.now().as_secs())
                        print("new subscription entry")
                    else:
                        update_subscription_sql_table(dvm_config.DB, event7001, recipient, subscriber, nwc,
                                                      cadence, overall_amount, start, end,
                                                      tier_dtag, zapsstr, recipe, isactivesubscription,
                                                      Timestamp.now().as_secs())
                        print("updated subscription entry")


                except Exception as e:
                    print(e)

            except Exception as e:
                print("Error in Subscriber " + str(e))

        self.client.handle_notifications(NotificationHandler())

        try:
            while True:
                time.sleep(60.0)
                subscriptions = get_all_subscriptions_from_sql_table(dvm_config.DB)

                for subscription in subscriptions:
                    if subscription.active:
                        if subscription.end < Timestamp.now().as_secs():
                            # We could directly zap, but let's make another check if our subscription expired
                            subscription_status = nip88_has_active_subscription(
                                PublicKey.parse(subscription.subscriber),
                                subscription.tier_dtag, self.client, subscription.recipent)

                            if not subscription_status["isActive"] or subscription_status["expires"]:
                                update_subscription_sql_table(dvm_config.DB, subscription.id,
                                                              subscription.recipent,
                                                              subscription.subscriber, subscription.nwc,
                                                              subscription.cadence, subscription.amount,
                                                              subscription.begin, subscription.end,
                                                              subscription.tier_dtag, subscription.zaps,
                                                              subscription.recipe,
                                                              False,
                                                              Timestamp.now().as_secs())
                            else:
                                zaps = json.loads(subscription.zaps)
                                success = pay_zap_split(subscription.nwc, subscription.amount, zaps)
                                if success:
                                    end = infer_subscription_end_time(Timestamp.now().as_secs(), subscription.cadence)
                                    recipe = make_subscription_zap_recipe(subscription.id, subscription.recipent,
                                                                          subscription.subscriber, subscription.begin,
                                                                          end, subscription.tier_dtag)
                                else:
                                    end = Timestamp.now().as_secs()
                                    recipe = subscription.recipe

                                update_subscription_sql_table(dvm_config.DB, subscription.id,
                                                              subscription.recipent,
                                                              subscription.subscriber, subscription.nwc,
                                                              subscription.cadence, subscription.amount,
                                                              subscription.begin, end,
                                                              subscription.tier_dtag, subscription.zaps, recipe,
                                                              success,
                                                              Timestamp.now().as_secs())
                                print("updated subscription entry")


                    else:
                        delete_threshold = 60 * 60 * 24 * 365
                        if subscription.cadence == "daily":
                            delete_threshold = 60 * 60 * 24 * 3  # After 3 days, delete the subscription, user can make a new one
                        elif subscription.cadence == "weekly":
                            delete_threshold = 60 * 60 * 24 * 21  # After 21 days, delete the subscription, user can make a new one
                        elif subscription.cadence == "monthly":
                            delete_threshold = 60 * 60 * 24 * 60  # After 60 days, delete the subscription, user can make a new one
                        elif subscription.cadence == "yearly":
                            delete_threshold = 60 * 60 * 24 * 500  # After 500 days, delete the subscription, user can make a new one

                        if subscription.end < (Timestamp.now().as_secs() - delete_threshold):
                            delete_from_subscription_sql_table(dvm_config.DB, subscription.id)
                            print("Delete expired subscription")

                print(str(Timestamp.now().as_secs()) + ": Checking " + str(len(subscriptions)) + " Subscription entries..")

        except KeyboardInterrupt:
            print('Stay weird!')
            os.kill(os.getpid(), signal.SIGTERM)
